/*  
    DeSpot - Conditional Temporal Despotting Filter for Avisynth 2.5

Filter functions

    Version 1.0  December 30, 2003
    with "sign", "fitluma", "blur" , "tsmooth" parameters added by Fizick     
    SSE branches is not modified now.

Copyright (C)2003 Alexander G. Balakhnin aka Fizick.
bag@hotmail.ru         http://bag.hotmail.ru
under the GNU General Public Licence version 2.

This filter is based on Conditional Temporal Median Filter for Avisynth 2.5
Version 0.93, September 27, 2003
Copyright (C) 2003 Kevin Atkinson (kevin.med@atkinson.dhs.org) under the GNU GPL version 2.
http://kevin.atkinson.dhs.org/temporal_median/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <new>
#include <string.h>
#include <stdlib.h>

#include "despot.hpp"

#if defined(SSE_GCC) || defined(SSE_INTEL)
#  define SSE
#endif

#ifdef SSE

#if defined(SSE_GCC)
#  include "sse-gcc.hpp"
#elif defined(SSE_INTEL)
#  include "sse-intel.hpp"
#endif

struct CWByteP
{
  const v8ubyte * d;
  CWByteP(const byte * d0) : d((const v8ubyte *)d0) {}
  CWByteP(ConstRow     d0) : d((const v8ubyte *)d0.data) {}
  const v8ubyte & operator[] (int i) const {return d[i];}
  void operator+= (int i) {d += i/8;}
};

struct MWByteP
{
  v8ubyte * d;
  MWByteP(byte     * d0) : d((v8ubyte *)d0) {}
  MWByteP(MutableRow d0) : d((v8ubyte *)d0.data) {}
  v8ubyte & operator[] (int i) const {return d[i];}
  void operator+= (int i) {d += i/8;}
};

#define MM_EMPTY mm_empty()

#else // ndef SSE

struct CWByteP
{
  const byte * d;
  CWByteP(const byte * d0) : d(d0) {}
  CWByteP(ConstRow     d0) : d(d0.data) {}
  const byte operator[] (int i) const {return d[i];}
  void operator+= (int i) {d += i;}
};

struct MWByteP
{
  byte * d;
  MWByteP(byte     * d0) : d(d0) {}
  MWByteP(MutableRow d0) : d(d0.data) {}
  byte & operator[] (int i) const {return d[i];}
  void operator+= (int i) {d += i;}
};

#define MM_EMPTY

#endif // SSE

extern "C"
void memzero(void * d0, size_t s)
{
  memset(d0, 0, s);
  //int * d = (int *)d0;
  //for (size_t i = 0; i != s/4; ++i)
  //  *d = 0;
}

//
//
//

static const char B_NOTHING = 0;
static const char B_BORDER = 1;
static const char B_LINK = 2;
static const char B_SIZE = 3;
struct Border {
  char p1; // boolean
  char what;
  short ly;
  short lx1;
  short lx2;
  union Data {
    Data(short x = 0, short y = 0)
      {b.x1 = x; b.x2 = x; b.y1 = y; b.y2 = y;}
    struct {short x1; short x2; short y1; short y2;} b;
    struct {short width; short height; bool is_noise;} s;
    Border * nowis;
  } data;
  Border(short ly0, short lx10) : p1(false), what(B_BORDER),
                                  ly(ly0), lx1(lx10), lx2(lx10),
                                  data(lx10, ly0) {}
};

size_t border_size = sizeof(Border);

void Borders::setup(size_t s)
{
  data = (Border *) malloc (sizeof(Border) * (s/2 + 1));
  end = data;
}

//
//
//

extern "C" REGPARM void combine_xs(Borders & bds, byte * noise, const Parms &);
extern "C" REGPARM void combine_ys(Borders & bds, const Parms &);
extern "C" REGPARM void to_size(Borders & bds, const Parms &);

extern "C"
void find_sizes(Borders & bds, byte * noise, const Parms & parms)
{
  combine_xs(bds, noise, parms);
  combine_ys(bds, parms);
  to_size(bds, parms);
}

extern "C"
void combine_xs(Borders & bds, byte * noise, const Parms & p)
{
  byte * l = noise;
  Border * c = 0;
  for (size_t y = 0; y != p.height; ++y) 
  {
    for (size_t x = 0; x != p.width; ++x)
    {
      if (l[x] > p.p2) {

        if (c == 0) {c = bds.end++; new (c)Border(y,x);}
        if (l[x] > p.p1) c->p1 = true;

      } else if (c != 0) {

        c->data.b.x2 = x - 1;
        c->lx2 = x - 1; c = 0;

      }
    }
    if (c != 0) {c->lx2 = p.width - 1; c = 0;}
    l += p.width;
  }
}

extern "C"
void combine_ys(Borders & bds, const Parms & p)
{
  bds.end->ly = p.height; // to avoid some special cases
  Border * c  = bds.data;
  Border * n = c + 1;
  while (n->ly == (n-1)->ly) ++n;
  Border * n0;

  bool c_has_next, n_has_next;
  
  while (n < bds.end) 
  {
    n0 = n;
    if (n->ly < c->ly + p.y_next) {++n; while (n->ly == (n-1)->ly) ++n;}

    if (c->ly + p.y_next == n->ly) { 
      do {
        if ( (c->lx1 <= n->lx1 && c->lx2 >= n->lx1 - 1) ||
             (c->lx1 >  n->lx1 && n->lx2 >= c->lx1 - 1)   )
        {
          Border * a = c; while (a->what == B_LINK) a = a->data.nowis;
          Border * b = n; while (b->what == B_LINK) b = b->data.nowis;
          if (a > b) {Border * t = a; a = b; b = t;} // this is important
          a->p1 = a->p1 || b->p1;
          a->data.b.x1 = min(a->data.b.x1, b->data.b.x1);
          a->data.b.x2 = max(a->data.b.x2, b->data.b.x2);
          a->data.b.y2 = n->ly;
          if (a != b) {
            b->what = B_LINK;
            b->data.nowis = a;
          }
        }

        c_has_next = c->ly == (c+1)->ly;
        n_has_next = n->ly == (n+1)->ly;
        if      (c->lx2 <= n->lx2 && c_has_next) ++c;
        else if (n_has_next)                     ++n;
        else if (c_has_next)                     ++c;
        else break;

      } while (true);

      ++c;
      ++n;

    } else {
      if (c->ly < n0->ly && n0->ly < n->ly) {
        c = n0;
      } else {
        c = n;
        ++n;
        while (n->ly == (n-1)->ly) ++n;
      }
    }
  }
}

extern "C"
void to_size(Borders & bds, const Parms & p)
{
  const short PH = p.pheight*p.y_next;
  for (Border * cur = bds.data; cur != bds.end; ++cur)
  {
    if (cur->what == B_BORDER) {
      Border::Data & b = cur->data;
      short w = b.b.x2 - b.b.x1 + 1;
      short h = b.b.y2 - b.b.y1 + 1;
      cur->what = B_SIZE; 
      b.s.is_noise = cur->p1 && w <= p.pwidth && h <= PH;
      b.s.width = w; 
      b.s.height = h;
    } else {
      //assert(cur->what == B_LINK);
      //assert(cur->data.nowis->what == B_SIZE);
      cur->what = B_SIZE;
      cur->data = cur->data.nowis->data;
    }
  }
}

extern "C"
void mark_noise(Borders & bds, byte * l, const Parms & p)
{
  memzero(l, p.size);
  Border * c = bds.data;
  bds.end->ly = p.height; // to avoid having to check for < bds.end
  for (int y = 0; y != (int) p.height; ++y)       // (int) added by Fizick
  {
    for (; c->ly == y; ++c) {
      if (c->data.s.is_noise) {
        byte * b = l + c->lx1;
        byte * end = l + c->lx2 + 1;
        for (; b != end; ++b) *b = 0xFF;
      }
    }
    l += p.width;
  }
}

//
// Motion Detection
//

extern "C"
void find_motion(ConstRow P, ConstRow C, byte * motion, 
                 const Parms & parms)
{{
  size_t width = parms.width;
  CWByteP p(P), c(C);
  MWByteP motion(motion);
#ifdef SSE
  width /= 8;
  v8ubyte thres; thres.set(parms.mthres+1);
  v8ubyte v_3; v_3.set(3);
#endif
  for (size_t y = 0; y != parms.height; ++y) {
    for (size_t x = 0; x != width; ++x) 
    {
#ifdef SSE
      v8ubyte mask = dist(p[x], c[x]) >= thres;
      motion[x] = mask & v_3;
#else
      if (abs(p[x] - c[x]) > parms.mthres) motion[x] = 3;
      else                                 motion[x] = 0;
#endif
    }
    p += P.pitch; c += C.pitch;
    motion += parms.width;
  }
  MM_EMPTY;
}}

extern "C"
void motion_denoise(byte * moving, const Parms & p)
{
  // This function is based on the motion map denoise code found in
  // Donald Graft's Smart Deinterlacer Filter.  I completely rewrote
  // it to so that its complexity is O(w*h) instead of
  // O(w*h*MWIDTH*MHEIGHT).

  const int w = p.width, h = p.height;
  const int w_8 = w / 8;

  byte fmoving[h*w];
  byte row[p.y_next][((w+64)/8)*8 + 8];

  const int MHEIGHT_2 = p.mheight/2;
  const int MWIDTH_2  = p.mwidth/2;
  const int Y_NEXT    = p.y_next;
  const int MPx3      = p.mp*3;

  // Erode
  memzero(row, sizeof(row));
  for (int y = -(MHEIGHT_2)*Y_NEXT; y < h; ++y)
  {
    byte * r = row[y & (Y_NEXT - 1)] + 32;
    int y1 = y - (MHEIGHT_2)*Y_NEXT - Y_NEXT, y2 = y + (MHEIGHT_2)*Y_NEXT;
#ifdef SSE
    v8ubyte * r_v = (v8ubyte *)r;
    if (y1 >= 0) {
      v8ubyte * r2_v = (v8ubyte *)(moving + w * y1);
      for (int x = 0; x != w_8; ++x)
        r_v[x] -= r2_v[x];}
    if (y2 < h) {
      v8ubyte * r2_v = (v8ubyte *)(moving + w * y2);
      for (int x = 0; x != w_8; ++x)
        r_v[x] += r2_v[x];}
#else
    if (y1 >= 0) {
      byte * r2 = moving + w * y1;
      for (int x = 0; x != w; ++x)
        r[x] -= r2[x];}
    if (y2 < h) {
      byte * r2 = moving + w * y2;
      for (int x = 0; x != w; ++x)
        r[x] += r2[x];}
#endif
    if (y >= 0) {
      byte * r2 = moving + w * y;
      byte * fw = fmoving + w * y;
      int total = 0;
      for (int x = 0; x != MWIDTH_2; ++x)
        total += r[x];
      for (int x = 0; x != w; ++x) {
        total -= r[x - MWIDTH_2 - 1];
        total += r[x + MWIDTH_2];
        fw[x] = r2[x] && total > MPx3 ? 1 : 0;}}
  }

  // Dilate  
  memzero(row, sizeof(row));
  for (int y = -(MHEIGHT_2)*Y_NEXT; y < h; ++y)
  {
    byte * r = row[y & (Y_NEXT - 1)] + 32;
    int y1 = y - (MHEIGHT_2)*Y_NEXT - Y_NEXT, y2 = y + (MHEIGHT_2)*Y_NEXT;
#ifdef SSE
    v8ubyte * r_v = (v8ubyte *)r;
    if (y1 >= 0) {
      v8ubyte * r2_v = (v8ubyte *)(fmoving + w * y1);
      for (int x = 0; x != w_8; ++x)
        r_v[x] -= r2_v[x];}
    if (y2 < h) {
      v8ubyte * r2_v = (v8ubyte *)(fmoving + w * y2);
      for (int x = 0; x != w_8; ++x)
        r_v[x] += r2_v[x];}
#else
    if (y1 >= 0) {
      byte * r2 = fmoving + w * y1;
      for (int x = 0; x != w; ++x)
        r[x] -= r2[x];}
    if (y2 < h) {
      byte * r2 = fmoving + w * y2;
      for (int x = 0; x != w; ++x)
        r[x] += r2[x];}
#endif
    if (y >= 0) {
      byte * fw = moving + w * y;
      int total = 0;
      for (int x = 0; x != MWIDTH_2; ++x)
        total += r[x];
      for (int x = 0; x != w; ++x) {
        total -= r[x - MWIDTH_2 - 1];
        total += r[x + MWIDTH_2];
        fw[x] = total ? 255 : 0;}}
  }
  MM_EMPTY;
}

//
// Noise Detection
//

extern "C"
void cond_median(ConstRow P, 
                 ConstRow C, byte * c_noise, byte * c_motion, 
                 ConstRow N, byte * n_motion, 
                 MutableRow O, const Parms & parms)
{{
  size_t width = parms.width;
  CWByteP p(P), c(C), n(N);
  MWByteP c_noise(c_noise), c_motion(c_motion), n_motion(n_motion);
  MWByteP o(O);
#ifdef SSE
  width /= 8;
#endif
  for (size_t y = 0; y != parms.height; ++y) {
    for (size_t x = 0; x != width; ++x) {
#ifdef SSE
      v8ubyte mask = c_motion[x] | n_motion[x];
      v8ubyte mn = min(p[x], n[x]);
      v8ubyte mx = max(p[x], n[x]);
      v8ubyte tmp = c[x];
      tmp = max(tmp, mn);
      tmp = min(tmp, mx);
      o[x] = if_else(mask, c[x], tmp);
#else
      o[x] = c[x];
      if (c_motion[x] || n_motion[x]) continue;
      byte mn = min(p[x], n[x]);
      byte mx = max(p[x], n[x]);
      o[x] = max(mn, o[x]);
      o[x] = min(mx, o[x]);
#endif
    }
    p += P.pitch; c += C.pitch; n += N.pitch;
    o += O.pitch;
    c_noise += parms.width; c_motion += parms.width;
    n_motion += parms.width;
  }
  MM_EMPTY;
}}

extern "C"
void find_outliers(ConstRow P, ConstRow C, ConstRow N, 
                   byte * o, const Parms & parms)
{{
  size_t width = parms.width;
  CWByteP p(P), c(C), n(N);
  MWByteP o(o);
//Sorry, SSE part is not coded by Fizick for present,
//    SSE branches is not modified now.
#ifdef SSE
  width /= 8;
#endif
  for (size_t y = 0; y != parms.height; ++y) {
    for (size_t x = 0; x != width; ++x) {
#ifdef SSE
      v8ubyte mn = min(p[x], n[x]);
      v8ubyte mx = max(p[x], n[x]);
      o[x] = ssub(mn, c[x]) | ssub(c[x], mx);
#else
      o[x] = 0;
      byte mn = min(p[x], n[x]);
      byte mx = max(p[x], n[x]);
//Original code by Kevin Atkinson:
//      if (c[x] < mn) o[x] = mn - c[x];
//      if (c[x] > mx) o[x] = c[x] - mx;
// Modes added by Fizick:
// Set some black-white mode of spots search, from sign parameter:
      switch (parms.sign){
      case 1:
//  sign=1 - only black (dark) spots and outliers 
        if (c[x] < mn) o[x] = mn - c[x];
        break;
      case -1:
//  sign=-1 - only white (light) spots and outliers 
        if (c[x] > mx) o[x] = c[x] - mx;
        break;
      case 2:
//  sign=2 - only black (dark) spots, any outliers 
        if (c[x] < mn) o[x] = mn - c[x];
        if (c[x] > mx) o[x] = min( (c[x] - mx), parms.p1 );
        break;
      case -2:
//  sign=-2 - only white (light) spots, any outliers 
        if (c[x] > mx) o[x] = c[x] - mx;
        if (c[x] < mn) o[x] = min( (mn - c[x]), parms.p1 );
        break;
      case 0:
      default:  
//  sign=0 - any spots and outliers (default)
        if (c[x] < mn) o[x] = mn - c[x];
        if (c[x] > mx) o[x] = c[x] - mx;
      }
#endif
    }
    p += P.pitch; c += C.pitch; n += N.pitch;
    o += parms.width;
  }
  MM_EMPTY;
}}


extern "C"
void mark_motion(ConstRow P, byte * p_noise,
                 ConstRow C, byte * c_noise, byte * c_motion,
                 const Parms & parms)
{{
  size_t width = parms.width;
  CWByteP p(P), c(C);
  MWByteP p_noise(p_noise), c_noise(c_noise), c_motion(c_motion);
#ifdef SSE
  width /= 8;
  v8ubyte thres; thres.set(parms.mthres+1);
  v8ubyte v_1; v_1.set(1);
  v8ubyte v_3; v_3.set(3);
  v8ubyte m1,m2,m3,m4;
#endif
  for (size_t y = 0; y != parms.height; ++y) {
    for (size_t x = 0; x != width; ++x) 
    {
#ifdef SSE
      v8ubyte d_mask = dist(p[x], c[x]) >= thres;
      v8ubyte n_mask = p_noise[x] | c_noise[x];
      c_motion[x] = d_mask & if_else(n_mask, v_1, v_3);
#else
      if (abs(p[x] - c[x]) > parms.mthres) {
        if (p_noise[x] || c_noise[x])  c_motion[x] = 1;
        else                           c_motion[x] = 3;
      } else                           c_motion[x] = 0;
#endif
    }
    p += P.pitch; c += C.pitch;
    p_noise += parms.width;
    c_noise += parms.width; c_motion += parms.width;
  }
  MM_EMPTY;
}}

//
//    1D  blur line near deleted spot borders.
//    Added by Fizick
//
extern "C"
void lineblur(MWByteP c_noise, MWByteP c_motion, MWByteP n_motion, MWByteP o, size_t width, const Parms & parms)
{
    int blur = parms.blur; // 1d blur radius
    byte b[20];	           // buffer
    switch(blur) {                
      case 0:                       // no blur
          break;
      case 1:
     	  for (size_t x = 2*blur; x < width-2*blur; ++x) {   // sub blur diameter to not exceed bounds
            if ( c_noise[x] && !c_motion[x] && !n_motion[x]) {      
              if (!c_noise[x-1]){            	// spot begins at x 
                b[0] = (o[x-2] + 2*o[x-1] + o[x])/4  ;         // make line (1D) blur
                b[1] = (o[x-1] + 2*o[x] + o[x+1])/4  ;
                o[x-1] = b[0];
                o[x] = b[1];
              }  
              else if (!c_noise[x+1]){          // spot ends at x
                b[0] = (o[x-1] + 2*o[x] + o[x+1])/4  ;
                b[1] = (o[x] + 2*o[x+1] + o[x+2])/4  ;
                o[x] = b[0];
                o[x+1] = b[1];
                } 
            }
          }       
          break;
      case 2:
      	  for (size_t x = 2*blur; x < width-2*blur; ++x) {   // sub blur diameter to not exceed bounds
            if ( c_noise[x] && !c_motion[x] && !n_motion[x]) {      
              if (!c_noise[x-1]){            	// spot begins at x 
                b[3] = (o[x-4] + 2*o[x-3] + 3*o[x-2] + 2*o[x-1]+ o[x])/9;
                b[4] = (o[x-3] + 2*o[x-2] + 3*o[x-1] + 2*o[x]+ o[x+1])/9;
                b[5] = (o[x-2] + 2*o[x-1] + 3*o[x] + 2*o[x+1]+ o[x+2])/9;
                b[6] = (o[x-1] + 2*o[x] + 3*o[x+1] + 2*o[x+2]+ o[x+3])/9;
                o[x-2] = b[3];
                o[x-1] = b[4];
                o[x] = b[5];
                o[x+1] = b[6];
              }  
              else if (!c_noise[x+1]){          // spot ends at x
                b[4] = (o[x-3] + 2*o[x-2] + 3*o[x-1] + 2*o[x]+ o[x+1])/9;
                b[5] = (o[x-2] + 2*o[x-1] + 3*o[x] + 2*o[x+1]+ o[x+2])/9;
                b[6] = (o[x-1] + 2*o[x] + 3*o[x+1] + 2*o[x+2]+ o[x+3])/9;
                b[7] = (o[x] + 2*o[x+1] + 3*o[x+2] + 2*o[x+3]+ o[x+4])/9;
                o[x-1] = b[4];
                o[x] = b[5];
                o[x+1] = b[6];
                o[x+2] = b[7];
              }  
            }
          }       
          break;
      case 3:
     	  for (size_t x = 2*blur; x < width-2*blur; ++x) {   // sub blur diameter to not exceed bounds
            if ( c_noise[x] && !c_motion[x] && !n_motion[x]) {      
              if (!c_noise[x-1]){            	// spot begins at x 
                b[2] = (o[x-6] + 2*o[x-5] + 3*o[x-4] + 4*o[x-3] + 3*o[x-2] + 2*o[x-1]+ o[x])/16;
                b[3] = (o[x-5] + 2*o[x-4] + 3*o[x-3] + 4*o[x-2] + 3*o[x-1] + 2*o[x]+ o[x+1])/16;
                b[4] = (o[x-4] + 2*o[x-3] + 3*o[x-2] + 4*o[x-1] + 3*o[x] + 2*o[x+1]+ o[x+2])/16;
                b[5] = (o[x-3] + 2*o[x-2] + 3*o[x-1] + 4*o[x] + 3*o[x+1] + 2*o[x+2]+ o[x+3])/16;
                b[6] = (o[x-2] + 2*o[x-1] + 3*o[x] + 4*o[x+1] + 3*o[x+2] + 2*o[x+3]+ o[x+4])/16;
                b[7] = (o[x-1] + 2*o[x] + 3*o[x+1] + 4*o[x+2] + 3*o[x+3] + 2*o[x+4]+ o[x+5])/16;
                o[x-3] = b[2];
                o[x-2] = b[3];
                o[x-1] = b[4];
                o[x] = b[5];
                o[x+1] = b[6];
                o[x+2] = b[7];
              }  
              else if (!c_noise[x+1]){          // spot ends at x
                b[3] = (o[x-5] + 2*o[x-4] + 3*o[x-3] + 4*o[x-2] + 3*o[x-1] + 2*o[x]+ o[x+1])/16;
                b[4] = (o[x-4] + 2*o[x-3] + 3*o[x-2] + 4*o[x-1] + 3*o[x] + 2*o[x+1]+ o[x+2])/16;
                b[5] = (o[x-3] + 2*o[x-2] + 3*o[x-1] + 4*o[x] + 3*o[x+1] + 2*o[x+2]+ o[x+3])/16;
                b[6] = (o[x-2] + 2*o[x-1] + 3*o[x] + 4*o[x+1] + 3*o[x+2] + 2*o[x+3]+ o[x+4])/16;
                b[7] = (o[x-1] + 2*o[x] + 3*o[x+1] + 4*o[x+2] + 3*o[x+3] + 2*o[x+4]+ o[x+5])/16;
                b[8] = (o[x] + 2*o[x+1] + 3*o[x+2] + 4*o[x+3] + 3*o[x+4] + 2*o[x+5]+ o[x+6])/16;
                o[x-2] = b[3];
                o[x-1] = b[4];
                o[x] = b[5];
                o[x+1] = b[6];
                o[x+2] = b[7];
                o[x+3] = b[8];
              }  
            }
          }       
          break;
      case 4:
      default:
     	  for (size_t x = 2*4; x < width-2*4; ++x) {   // sub blur diameter to not exceed bounds
            if ( c_noise[x] && !c_motion[x] && !n_motion[x]) {      
              if (!c_noise[x-1]){            	// spot begins at x 
                b[1] = (o[x-8] + 2*o[x-7] + 3*o[x-6] + 4*o[x-5] + 5*o[x-4] + 4*o[x-3]+ 3*o[x-2] + 2*o[x-1]+ o[x])/25;
                b[2] = (o[x-7] + 2*o[x-6] + 3*o[x-5] + 4*o[x-4] + 5*o[x-3] + 4*o[x-2]+ 3*o[x-1] + 2*o[x]+ o[x+1])/25;
                b[3] = (o[x-6] + 2*o[x-5] + 3*o[x-4] + 4*o[x-3] + 5*o[x-2] + 4*o[x-11]+ 3*o[x] + 2*o[x+1]+ o[x+2])/25;
                b[4] = (o[x-5] + 2*o[x-4] + 3*o[x-3] + 4*o[x-2] + 5*o[x-1] + 4*o[x]+ 3*o[x+1] + 2*o[x+2]+ o[x+3])/25;
                b[5] = (o[x-4] + 2*o[x-3] + 3*o[x-2] + 4*o[x-1] + 5*o[x] + 4*o[x+1]+ 3*o[x+2] + 2*o[x+3]+ o[x+4])/25;
                b[6] = (o[x-3] + 2*o[x-2] + 3*o[x-1] + 4*o[x] + 5*o[x+1] + 4*o[x+2]+ 3*o[x+3] + 2*o[x+4]+ o[x+5])/25;
                b[7] = (o[x-2] + 2*o[x-1] + 3*o[x] + 4*o[x+1] + 5*o[x+2] + 4*o[x+3]+ 3*o[x+4] + 2*o[x+5]+ o[x+6])/25;
                b[8] = (o[x-1] + 2*o[x] + 3*o[x+1] + 4*o[x+2] + 5*o[x+3] + 4*o[x+4]+ 3*o[x+5] + 2*o[x+6]+ o[x+7])/25;
                b[9] = (o[x] + 2*o[x+1] + 3*o[x+2] + 4*o[x+3] + 5*o[x+4] + 4*o[x+5]+ 3*o[x+6] + 2*o[x+7]+ o[x+8])/25;
                o[x-4] = b[1];
                o[x-3] = b[2];
                o[x-2] = b[3];
                o[x-1] = b[4];
                o[x] = b[5];
                o[x+1] = b[6];
                o[x+2] = b[7];
                o[x+3] = b[8];
              }  
              else if (!c_noise[x+1]){          // spot ends at x
                b[2] = (o[x-7] + 2*o[x-6] + 3*o[x-5] + 4*o[x-4] + 5*o[x-3] + 4*o[x-2]+ 3*o[x-1] + 2*o[x]+ o[x+1])/25;
                b[3] = (o[x-6] + 2*o[x-5] + 3*o[x-4] + 4*o[x-3] + 5*o[x-2] + 4*o[x-11]+ 3*o[x] + 2*o[x+1]+ o[x+2])/25;
                b[4] = (o[x-5] + 2*o[x-4] + 3*o[x-3] + 4*o[x-2] + 5*o[x-1] + 4*o[x]+ 3*o[x+1] + 2*o[x+2]+ o[x+3])/25;
                b[5] = (o[x-4] + 2*o[x-3] + 3*o[x-2] + 4*o[x-1] + 5*o[x] + 4*o[x+1]+ 3*o[x+2] + 2*o[x+3]+ o[x+4])/25;
                b[6] = (o[x-3] + 2*o[x-2] + 3*o[x-1] + 4*o[x] + 5*o[x+1] + 4*o[x+2]+ 3*o[x+3] + 2*o[x+4]+ o[x+5])/25;
                b[7] = (o[x-2] + 2*o[x-1] + 3*o[x] + 4*o[x+1] + 5*o[x+2] + 4*o[x+3]+ 3*o[x+4] + 2*o[x+5]+ o[x+6])/25;
                b[8] = (o[x-1] + 2*o[x] + 3*o[x+1] + 4*o[x+2] + 5*o[x+3] + 4*o[x+4]+ 3*o[x+5] + 2*o[x+6]+ o[x+7])/25;
                b[9] = (o[x] + 2*o[x+1] + 3*o[x+2] + 4*o[x+3] + 5*o[x+4] + 4*o[x+5]+ 3*o[x+6] + 2*o[x+7]+ o[x+8])/25;
                o[x-3] = b[2];
                o[x-2] = b[3];
                o[x-1] = b[4];
                o[x] = b[5];
                o[x+1] = b[6];
                o[x+2] = b[7];
                o[x+3] = b[8];
                o[x+4] = b[9];
              }  
            }
          }       
    }	
	
}	

//
//
//

extern "C"
void remove_outliers(ConstRow P,
                     ConstRow C, byte * c_noise, byte * c_motion,
                     ConstRow N, byte * n_motion,
                     MutableRow O, 
                     const Parms & parms)
{{
  size_t width = parms.width;
  CWByteP p(P), c(C), n(N);
  MWByteP c_noise(c_noise), c_motion(c_motion), n_motion(n_motion);
  MWByteP o(O);
  long lumac,lumap,luman,lumam;
  int lumao, lumad=0;
  int sigmaii2, sigmass2, mean, tsmoothed;
  int sigmann2 = (parms.tsmooth)*(parms.tsmooth);    // approximate value of noise variance for Wiener temporal smoothing

//Sorry, SSE part is not coded by Fizick for present,
//    SSE branches is not modified now.
#ifdef SSE
  width /= 8;
#endif

  for (size_t y = 0; y != parms.height; ++y) {   

    if (parms.fitluma) {                // fitluma added by Fizick
        lumac = 0;                        
        lumap = 0;
        luman = 0;
        for (size_t x = 0; x != width; ++x) {   
            lumac += c[x];               //  prepare line lumas
            lumap += p[x];
            luman += n[x];    
        }
        lumac /= width;
        lumap /= width;
        luman /= width; 
        lumam = (luman+lumap)/2;        // mean luma of prev and next frame
        lumad = lumac-lumam;            //  luma delta
    }
    
    for (size_t x = 0; x != width; ++x) 
    {
#ifdef SSE
      v8ubyte mask = andn(c_motion[x] | n_motion[x], c_noise[x]);
      o[x] = if_else(mask, avg(p[x], n[x]), c[x]);
#else
      if (c_noise[x] && !c_motion[x] && !n_motion[x]) {      
      	
        o[x] = (p[x] + n[x] + 1)/2;
      	
      	if (parms.fitluma) {                // fitluma added by Fizick
        	lumao = o[x]+ lumad;        // luma correction  of deleted spot place
        	if (lumao>255) o[x]=255;   // check if not overflow
        	else if (lumao<0) o[x]=0;
        	     else  o[x] = lumao;
        }
      }       
      else
      {
        o[x] = c[x];
      }
#endif
    }

    if (parms.blur) {   // if parameter blur is set, then blur line near spot borders  // added by Fizick:
           lineblur(c_noise, c_motion, n_motion, o, width, parms);
    }       

    for (size_t x = 0; x != width; ++x) // temporal smoothing must be after borders blur!  // added by Fizick:
    {
      if ((sigmann2 > 0) && !c_noise[x] && !c_motion[x] && !n_motion[x]) {    
// Wiener temporal smoothing  for static points, except noise(spot), and motion  
         mean = (p[x]+o[x]+n[x])/3;        // mean value
         sigmass2 = ((mean-p[x])*(mean-p[x])+(mean-o[x])*(mean-o[x])+(mean-n[x])*(mean-n[x]))/3;    // variance of source
         sigmaii2 = sigmass2 - sigmann2;   // predicted variance of cleaned signal = var.source - var.noise
         if (sigmaii2<0)  sigmaii2 = 0;    // but not negative
         tsmoothed = (sigmaii2*o[x]+sigmann2*mean)/(sigmaii2+sigmann2);   // Wiener filter
         if (tsmoothed>255)  tsmoothed = 255;
         o[x] = tsmoothed;
      }                    
    }

    p += P.pitch; c += C.pitch; n += N.pitch;
    o += O.pitch;
    c_noise += parms.width; c_motion += parms.width;
    n_motion += parms.width;
  }
  MM_EMPTY;
}}

extern "C"
void mark_outliers(ConstRow P,
                   ConstRow C, byte * c_noise, byte * c_motion,
                   ConstRow N, byte * n_motion,
                   MutableRow O, 
                   const Parms & parms)
{
  const byte * p = P.data, * c = C.data, * n = N.data;
  byte * o = O.data; 
  for (size_t y = 0; y != parms.height; ++y) {
    for (size_t x = 0; x != parms.width; ++x) {
      if (c_noise[x] && !(c_motion[x] || n_motion[x]))
        o[x] = parms.mark_v;
      else
        o[x] = c[x];
    }
    p += P.pitch; c += C.pitch; n += N.pitch;
    o += O.pitch;
    c_noise += parms.width; c_motion += parms.width;
    n_motion += parms.width;
  }
}

extern "C"
void map_outliers(ConstRow P,
                  ConstRow C, byte * c_noise, byte * c_motion,
                  ConstRow N, byte * n_motion,
                  MutableRow O, 
                  const Parms & parms)
{
  const byte * p = P.data, * c = C.data, * n = N.data;
  byte * o = O.data; 
  for (size_t y = 0; y != parms.height; ++y) {
    for (size_t x = 0; x != parms.width; ++x) {
      if (c_noise && c_noise[x]) {
        if (c_motion[x] || n_motion[x])
          o[x] = 159;
        else
          o[x] = 255;
      } else {
        if (c_motion[x]) 
          o[x] = 95;
        else if (n_motion[x]) 
          o[x] = 63;
        else
          o[x] = 0;
      }
    }
    p += P.pitch; c += C.pitch; n += N.pitch;
    o += O.pitch;
    if (c_noise)
      c_noise += parms.width; 
    c_motion += parms.width;
    n_motion += parms.width;
  }
}
